# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import *
from ...ops import *
from ..VT import *


def make_idx_generator(n=0):
    def next_idx():
        nonlocal n
        n = n + 1
        return n
    return next_idx


def is_same_et(et1, et2):
    return without_absorbed(et1) == without_absorbed(et2)

def get_et_id(et):
    assert len(absorbed(et)) == 1, f"{et} didn't have excatly a single absorbed arg!"
    return absorbed(et)[0]

@func
def create_object(d: Dict, generator, idx_to_obj) -> Entity:

    def mapping(v):
        if isinstance(v, dict):
            return create_object(v, generator, idx_to_obj)
        elif isinstance(v, list):
            return [mapping(x) for x in v]
        elif isinstance(v, set):
            return {mapping(x) for x in v}
        else:
            return v
    idx = str(generator())
    if "type" in d:
        obj = ET(d['type'])[idx]
    else:
        obj = ET.ZEF_Unknown[idx]

    d = dict(d | items | map[lambda t: (t[0] , mapping(t[1]))] | collect)
    obj = obj(**d)
    idx_to_obj[idx] = obj
    return obj




def match_rule(rule, pattern):
    assert (isinstance(rule, list) or isinstance(rule, tuple)) and (isinstance(pattern, list) or isinstance(pattern, tuple))
    z_pos = - 1
    for i, (p_e,x_e) in enumerate(zip(rule, pattern)): 
        if p_e == Z:
            z_pos = i

        elif is_a(p_e, ET) and is_a(x_e, ET):
            if not is_same_et(p_e, x_e): return -1

        elif p_e != x_e and p_e != Any :
            return -1

    return z_pos



def flatten_object(obj: Entity) -> list[list[tuple, tuple]]:
    """
    Takes an object of type EntityValueInstance_ and flattens it out to pairs of tuples ((source, rt, target), (src_id?, None, trgt_id?)).
    It recursively calls itself if at any level a non-terminal object is found.
    """
    from zef.core.patching import EntityValueInstance_
    
    @func
    def flatten_to_tuples_with_ids(obj, rt, trgt) -> tuple:
        def generate_type_and_id_maybe(obj):
            return (obj._entity_type, get_et_id(obj._entity_type)) if isinstance(obj, EntityValueInstance_) else (obj, None)
        
        return ((obj, rt, trgt) | map[generate_type_and_id_maybe] | zip | collect) # -> # [(obj, t[0], t[1]), (id?, None, None)]

    @func
    def mapping(rt_and_trgt, obj):
        rt, trgt = rt_and_trgt
        if isinstance(trgt, EntityValueInstance_):
            flattened = flatten_object(trgt)
            return [flatten_to_tuples_with_ids(obj, rt, trgt), *flattened]
        
        elif isinstance(trgt, (list, set)):
            flattened = trgt | map[flatten_object] | concat | collect
            return [*(trgt | map[lambda trgt: flatten_to_tuples_with_ids(obj, rt, trgt)] | collect), *flattened]

        else:
            return [flatten_to_tuples_with_ids(obj, rt, trgt)]

    
    if not isinstance(obj, EntityValueInstance_): return []
    return obj._kwargs | items | map[mapping[obj]] | concat | collect


@func
def resolve_unknown(obj: Entity, rules: List, mapping: Dict) -> Entity:
    # For each path tries to match it against the rules set. If there is a match, the enity_type of the matched object is replaced.
    patterns_and_ids = flatten_object(obj) 
    for pattern, ids in patterns_and_ids:
        for rule, replacement in rules:
            Z_pos = match_rule(rule, pattern)
            # If there is a match and the object is not already of the correct type, the object is replaced.
            if Z_pos != -1 and mapping.get(ids[Z_pos], None) and is_same_et(mapping[ids[Z_pos]]._entity_type, ET.ZEF_Unknown):
                mapping[ids[Z_pos]]._entity_type = replacement[ids[Z_pos]]
                return obj

    return None


def validate_rules(rules: List) -> Bool:
    for rule,_ in rules:
        assert (rule | filter[lambda x: x == Z]  | length | equals[1] | collect), f"{rule} must contain exactly one Z!"
    return True


def infer_types(o: Dict|List, rules: List) -> Entity:
    """
    Given a dictionary (or a list of dicts) defining an object and a set of rules. The initial object with prefilled unknowns is iterated
    on until rules no longer match. The final object is returned.
    """
    validate_rules(rules)
    idx_to_obj = {}
    idx_generator = make_idx_generator()
    def create_and_resolve_single_dict(d: Dict):
        assert isinstance(d, dict), f"Expected a dict, got {d}"
        obj = create_object(d, idx_generator, idx_to_obj)
        LazyValue(obj) | iterate[resolve_unknown[rules][idx_to_obj]] | take_while[lambda x: x]  | collect
        return obj
    
    if isinstance(o, list):
        return o | map[create_and_resolve_single_dict] | collect
    elif isinstance(o, dict):
        return create_and_resolve_single_dict(o)
    else:
        raise Exception("Input must be a dictionary or a list of dictionaries.")

@func
def create_idx_to_obj_d(obj, idx_to_obj): 
    """
    Pure function that traverse the object and creates a dictionary mapping the internal ids of the objects to the objects themselves.
    """
    from zef.core.patching import EntityValueInstance_
    def traverse_nested(v):
        if isinstance(v, EntityValueInstance_):
            create_idx_to_obj_d(v, idx_to_obj)
        elif isinstance(v, (list, set)):
            [create_idx_to_obj_d(x, idx_to_obj) for x in v]

    if not isinstance(obj, EntityValueInstance_): return
    
    idx_to_obj[get_et_id(obj._entity_type)] = obj
    obj._kwargs | items | map[lambda t: traverse_nested(t[1])] | collect
   
    return idx_to_obj


@func
def identify_and_merge_step(obj, idx_to_obj, identification_rules):
    def retrieve_obj(et):
        return idx_to_obj[get_et_id(et)]
    
    def merge_identities(obj, other):
        # Remove overlapping obj from the dict
        idx_to_obj.pop(get_et_id(other._entity_type))
        
        # Overwrite the internal id of the other object with the internal id of the obj
        other._entity_type = obj._entity_type

        # Reset the overlapping object's internal dict
        other._kwargs = {}

    def try_identification(et, o1, o2):
        try:
            return identification_rules[et](o1, o2)
        except:
            return False
    if isinstance(obj, list):
        groups = obj | map[flatten_object] | concat | map[first | first] | func[set] | group_by[without_absorbed] | collect
    else:
        groups = flatten_object(obj) | map[first | first] | func[set] | group_by[without_absorbed] | collect
    for et, group in groups:
        if len(group) < 2: continue
        for o1, o2 in zip(group, group[1:]):
            o1, o2 = retrieve_obj(o1), retrieve_obj(o2)
            if try_identification(et, o1, o2):
                merge_identities(o1, o2)
                return obj
    
    return None


def deduplicate(obj_or_list, identification_rules: Dict) -> List[object]:
    idx_to_obj = {}
    if isinstance(obj_or_list, list):
        obj_or_list | for_each[create_idx_to_obj_d[idx_to_obj]]
    else:
        create_idx_to_obj_d(obj_or_list, idx_to_obj)
    LazyValue(obj_or_list) | iterate[identify_and_merge_step[idx_to_obj][identification_rules]]  | take_while[lambda x: x]  | collect
    return list(idx_to_obj.values())


# TODO merge with create_idx_to_obj_d
@func 
def generate_id_to_objs(obj, idx_to_objs):
    from zef.core.patching import EntityValueInstance_
    def traverse_nested(v):
        if isinstance(v, EntityValueInstance_):
            generate_id_to_objs(v, idx_to_objs)
        elif isinstance(v, (list, set)):
            [generate_id_to_objs(x, idx_to_objs) for x in v]

    if not isinstance(obj, EntityValueInstance_): return
    
    idx_to_objs[get_et_id(obj._entity_type)].add(obj)
    obj._kwargs | items | map[lambda t: traverse_nested(t[1])] | collect
   
    return idx_to_objs


def match_with_entity_and_replace_step(obj_list, idx_to_objs, identification_rules, gs):
    def try_retrieve_entity(obj):
        et = without_absorbed(obj._entity_type)
        try:
            return identification_rules[et](obj, gs)
        except:
            return None

    def replace_all_in_d(obj, ent):
        @func
        def over_write_obj(o, new_id): 
            o._entity_type = without_absorbed(o._entity_type)[new_id]
            o._kwargs = {}
            
        obj_id = get_et_id(obj._entity_type)
        entity_id = str(uid(ent))
        idx_to_objs[obj_id] | for_each[over_write_obj[entity_id]]

    for obj in obj_list:
        ent = try_retrieve_entity(obj)
        if ent: replace_all_in_d(obj, ent)



def identify_entities(obj_list: List[object], entity_identification_rules: Dict, gs: GraphSlice) -> List[object]:
    from collections import defaultdict
    idx_to_objs = defaultdict(set)
    obj_list | for_each[generate_id_to_objs[idx_to_objs]]
    match_with_entity_and_replace_step(obj_list, idx_to_objs, entity_identification_rules, gs)
    return obj_list


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import make_VT, insert_VT
from ... import pyzef

EZefRef = make_VT('EZefRef', pytype=pyzef.main.EZefRef)
ZefRef = make_VT('ZefRef', pytype=pyzef.main.ZefRef)
insert_VT("BlobPtr", ZefRef | EZefRef)
make_VT('GraphRef', pytype=pyzef.main.GraphRef)
make_VT('Time', pytype=pyzef.main.Time)

BaseUID = make_VT('BaseUID', pytype=pyzef.internals.BaseUID)
EternalUID = make_VT('EternalUID', pytype=pyzef.internals.EternalUID)
ZefRefUID = make_VT('ZefRefUID', pytype=pyzef.internals.ZefRefUID)
insert_VT("UID", (BaseUID | ZefRefUID | EternalUID))

make_VT('QuantityInt', pytype=pyzef.main.QuantityInt)
make_VT('QuantityFloat', pytype=pyzef.main.QuantityFloat)
make_VT('Enum', pytype=pyzef.main.ZefEnumValue)

def graph_ctor(*args, **kwargs):
    # Go to experimental load path for cases it supports
    # TODO: Uncomment this once event loop issue with IPython is figured out.
    # if len(kwargs) == 0 and len(args) == 1 and type(args[0]) == str:
    #     from ...experimental.repl_interface import load_graph
    #     return load_graph(args[0])
    # else:
    from ...core import internals
    return internals.Graph(*args, **kwargs)

Graph = make_VT('Graph', constructor_func=graph_ctor, pytype=pyzef.main.Graph)
insert_VT('DB', Graph)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import Effect, FX
from .._ops import *
from ..zef_functions import func
from ..VT import *
from ..logger import log


#-------------------------------------------------------------
#---------------------Resolvers-------------------------------
#-------------------------------------------------------------
# TODO: Add Better Error Reporting

@func
def graphs(query_args):
    from zef import pyzef
    uids = pyzef.internals.list_graph_manager_uids()
    return (
        uids
        | map[lambda uid: {'labels': [str(uid)], 'id': str(uid)}]
        | collect
    )
    return (
        globals() 
        | items     
        | filter[lambda kv: is_a(kv[1], Graph)]
        | map[lambda kv: {'labels': [kv[0]], 'id': str(uid(kv[1]))}]
        | collect
    )

@func
def entity_types(query_args):
    graph_id = query_args.get('graphID', None)
    if not graph_id: return None
    return list(Graph(graph_id) | now | all[ET] | map[rae_type] | func[set] | collect)


def make_cellet_value(obj):
    if has_out(obj, RT.Name):
        label = value(obj | Out[RT.Name])
    elif has_out(obj, RT.Label):
        label = value(obj | Out[RT.Label])
    else:
        label = None
        
    return {
            "type": "CellET",
            "id": uid(obj),
            "label": label,
            "etType": str(rae_type(obj)),
        }

def make_individual_value(obj):
   if is_a(obj, Entity): return make_cellet_value(obj)

   types = {AET.String: "CellString", AET.Float: "CellFloat", AET.Bool: "CellBoolean", AET.Int: "CellInt"}
      
   return {
      "type": types.get(rae_type(obj), "CellZef"),
      "id": str(uid(obj)),
      "value": obj | attempt[value][str(rae_type(obj))] | collect ,
   }

def make_cell(obj, rt):
   trgts = obj | Outs[rt] | collect
   if not trgts: 
      return None
   elif len(trgts) == 1:
      return make_individual_value(trgts[0])
   else:
      return {
         "type": "CellList",
         "id": f"{uid(obj)}_{rt}_{len(trgts)}",
         "value": [make_individual_value(obj) for obj in trgts]
      }

def make_row(obj, connected_rels):
   return {
      "cells": [make_cell(obj, rt) for rt in connected_rels]
   }


@func
def make_column(column):
   return {
      "header": str(column)
   }

def make_columns(g, entity_type):
   # TODO Add sorting of columns
   z = delegate_of(entity_type, g)
   connected_rels = z | out_rels[RT] | map[rae_type] | collect
   return connected_rels, connected_rels | map[make_column] | collect

def make_table_return(g, entity_type, entities):
   connected_rels, columns = make_columns(g, entity_type)
   return {
      # 'id': f"{graph_id}_{entity_type}",
      'columns': columns,
      'rows': [make_row(obj, connected_rels) for obj in entities],
   }

@func
def entity_table(query_args):
   graph_id = query_args.get('graphID', None)
   et_type  = query_args.get('entityType', None)
   if not graph_id or not et_type: return None

   et_type = ET(et_type)
   limit  = query_args.get('limit', 10)
   g = Graph(graph_id)
   return make_table_return(g, et_type, list(g | now |  all[et_type] | take[limit] | collect))

@func
def single_entity(query_args):
   graph_id = query_args.get('graphID', None)
   et_id  = query_args.get('entityID', None)
   if not graph_id or not et_id: return None

   g = Graph(graph_id)
   entity = g[et_id]
   if not is_a(entity, Entity): return None
   et_type = rae_type(entity)

   return make_table_return(g, et_type, [entity])

@func
def cell_interface_resolver(obj, *_):
   return obj.get('type', "CellString")


@func
def assign_value_general(query_args, aet_type):
    graphID = query_args['ids']['graphID']
    aetID = query_args['ids']['aetID']

    g = Graph(graphID)
    aet = g[aetID]

    if not is_a(aet, AttributeEntity): 
      log.error(f"Cannot assign value to {aetID} as it is not an AET")
      return False 

    if not is_a(aet, aet_type): 
      log.error(f"Cannot assign value to {aetID} as it is not an {aet_type}")
      return False

    try:
      aet_value = query_args['value']
      aet | assign[aet_value] | g | run
      
      return True
    except Exception as e:
      log.error(f"Failed to assign value to {aetID}. {e}")
      return False

@func
def assign_value_string(query_args):
    return assign_value_general(query_args, AET.String)

@func
def assign_value_float(query_args):
    return assign_value_general(query_args, AET.Float)

@func
def assign_value_int(query_args):
    return assign_value_general(query_args, AET.Int)

@func
def assign_value_bool(query_args):
    return assign_value_general(query_args, AET.Bool)


def value_of_aet_at_tx(aet, tx) -> str:
    try:
        if tx is now:
            tx = Graph(aet) | now | collect
        if BT(aet) == BT.ATTRIBUTE_ENTITY_NODE:
            val = aet | value[tx] | collect 
            return '' if val is None else str(val)
        else:
            return ''
    except:
        return ''

@func
def entity_events(query_args):
    graph_id = query_args.get('graphID', None)
    g = Graph(graph_id)

    rae_id  = query_args.get('raeID', None)
    zr      = g[rae_id]

    @func
    def events_for_rt(rt):
        zr_or_uzr = target(rt)
        uzr = to_ezefref(zr_or_uzr)

        low_lvl_mapping = {
            BT.ATOMIC_VALUE_ASSIGNMENT_EDGE: "ValueAssignedEvent",
            BT.ATTRIBUTE_VALUE_ASSIGNMENT_EDGE: "ValueAssignedEvent",
            BT.INSTANTIATION_EDGE: "InstantiatedEvent",
            BT.TERMINATION_EDGE: "TerminatedEvent",
        }

        def construct_event(edge):
            value_maybe = ""
            if BT(edge) in {BT.ATOMIC_VALUE_ASSIGNMENT_EDGE, BT.ATTRIBUTE_VALUE_ASSIGNMENT_EDGE}: 
                value_maybe = value_of_aet_at_tx(uzr, edge | source | collect)
            
            uzr_tx = source(edge)
            transaction = create_tx(uzr_tx)
            return create_event(low_lvl_mapping[BT(edge)], str(rae_type(rt)), transaction, value_maybe)

        rel_ent_inst_edge = uzr | in_rel[BT.RAE_INSTANCE_EDGE]
        return rel_ent_inst_edge | in_rels[BT] | map[construct_event] | collect
    
    return zr | out_rels | map[events_for_rt] | concat | sort[lambda d: d['transaction']['txTimestamp']] | collect


def create_tx(tx):
    return {
        "id": str(uid(tx)),
        "txTimestamp": time(tx).seconds_since_1970,
        "txIndex": graph_slice_index(to_graph_slice(tx)),
    }

def create_event(event_type, field_name, transaction, assignedValue=""):
    return { 
        "type": event_type,
        "transaction": transaction,
        "fieldName": field_name,
        "value": assignedValue,
    }

@func
def event_interface_resolver(obj, *_):
   return obj.get('type', "InstantiatedEvent")

#-------------------------------------------------------------
#-------------------Schema String-------------------------------
#-------------------------------------------------------------
simple_schema = """
type Query {
    graphs: [Graph]
    entityTypes(graphID: ID!): [String]
    entityTable(graphID: ID!, entityType: String!, limit: Int): Table
    entity(graphID: ID!, entityID: ID!): Table
    events(graphID: ID!, raeID: ID!): [Event]
}

type Mutation {
    assignValueString(ids: AssignValueIDs!, value: String!): Boolean
    assignValueFloat( ids: AssignValueIDs!, value: Float!): Boolean
    assignValueInt(   ids: AssignValueIDs!, value: Int!): Boolean
    assignValueBool(  ids: AssignValueIDs!, value: Boolean!): Boolean
}

input AssignValueIDs {
    graphID: ID!
    aetID: ID!
}

type Graph {
    id: ID!
    labels: [String]
}

type Table {
columns: [Column]
rows: [Row]
}

type Column {
header: String
}

type Row {
cells: [Cell]
}

interface Cell {
id: ID!
}

type CellInt implements Cell{
id: ID!
value: Int
}

type CellString implements Cell{
id: ID!
value: String
}

type CellFloat implements Cell{
id: ID!
value: Float
}

type CellBoolean implements Cell{
id: ID!
value: Boolean
}

type CellET implements Cell{
id: ID!
label: String
etType: String
}

type CellZef implements Cell{
id: ID!
value: String
}

type CellList implements Cell{
id: ID!
value: [Cell]
}

type Transaction {
    id: ID!
    txTimestamp: Float!
    txIndex: Int!
}

interface Event {
    transaction: Transaction!
    fieldName: String!
}

type InstantiatedEvent implements Event {
    transaction: Transaction!
    fieldName: String!
}

type TerminatedEvent implements Event {
    transaction: Transaction!
    fieldName: String!
}

type ValueAssignedEvent implements Event {
    transaction: Transaction!
    fieldName: String!
    value: String
}
"""

def create_schema_dict(simple_schema):
   from ...graphql import generate_schema_dict, fill_types_default_resolvers

   schema_dict = generate_schema_dict(simple_schema)
   schema_dict = fill_types_default_resolvers(schema_dict, lambda field_name: get[field_name])
   schema_dict = (
      schema_dict 
      | insert_in[('_Types', 'Query', 'graphs', 'resolver')][graphs] 
      | insert_in[('_Types', 'Query', 'entityTypes', 'resolver')][entity_types] 
      | insert_in[('_Types', 'Query', 'entityTable', 'resolver')][entity_table] 
      | insert_in[('_Types', 'Query', 'entity', 'resolver')][single_entity] 
      | insert_in[('_Types', 'Query', 'events', 'resolver')][entity_events]
      | insert_in[('_Types', 'Mutation', 'assignValueString', 'resolver')][assign_value_string]
      | insert_in[('_Types', 'Mutation', 'assignValueFloat', 'resolver')][assign_value_float]
      | insert_in[('_Types', 'Mutation', 'assignValueInt', 'resolver')][assign_value_int]
      | insert_in[('_Types', 'Mutation', 'assignValueBool', 'resolver')][assign_value_bool]
      | insert_in[('_Interfaces', 'Cell', '_interface_resolver')][cell_interface_resolver] 
      | insert_in[('_Interfaces', 'Event', '_interface_resolver')][event_interface_resolver] 
      | collect
   )

   return schema_dict

def studio_start_server_handler(eff: Dict):
   """Example
   {
      "type": FX.Studio.StartServer,
   } | run
   """
   def open_browser(port):
      studio_url = f"https://studio.zefhub.io/?endpoint=http://localhost:{port}/graphql"
      log.info(f"Started Zef Studio at {studio_url}")
      import webbrowser
      webbrowser.open(studio_url)

   import random
   from .. import internals
   g_process = internals.get_local_process_graph()

   schema_dict = create_schema_dict(simple_schema)

   trials = 5
   while trials:
      random_port = random.randint(10000, 30000)
      try:
         http_r = {
            "type": FX.GraphQL.StartServer,
            "schema_dict" : schema_dict,
            "g" :  Graph(),
            "port" :  random_port, 
            "path" :  "/graphql", 
         } | run

         server_zr = now(g_process[http_r['server_uuid']])
         
         ET.ZefFXService(**{
            "type": Val(eff['type']),
            'owns': server_zr,
         }) | g_process | run # TODO Perform that in a flat way by directly calling into the FX handler

         open_browser(random_port)
         return http_r
      except Exception as e:
         trials -= 1
         if trials: continue
         raise e


def studio_stop_server_handler(eff: Dict):
    return run({**eff, 'type': FX.HTTP.StopServer})

